/*
 * This is SSSim: the Simple & Stupid Simulator
 *
 *  Copyright (c) 2008 Luca Abeni
 *  Copyright (c) 2008 Csaba Kiraly
 *
 *  This is free software; see gpl-3.0.txt
 */

#include <limits.h>
#include <stdlib.h>
#include <stdio.h>

#include "config.h"
#include "core.h"
#include "sched.h"

FILE *trace;    //trace file
static struct peer source;

#define SCHED __attribute__ ((unused)) static inline

///==========================================================
///     helper functions
///==========================================================
//avoiding duplicate transactions
static int needs(int k, struct peer *p)
{
#ifdef NO_BUFFER
  return (!p->chunks[k] && !p->recv_chunks[k]);
#else
  register unsigned int idx = k & p->buf_mask;

  return ((!p->chunks[idx] || (p->chunks[idx]->chunk != k)) && !p->recv_chunks[idx]);
#endif
}

//LUCA style utility
static int utility(struct peer *p, int t)
{
  int i, c, u;

  u = 0;
  for (c = win_left(p, t); c < win_right(p, t); c++) {
    if (p->recv_chunks[win_idx(p, c)] || p->chunks[win_idx(p, c)]) {
      for (i = 0; i < p->neigh_size; i++) {
        u += needs(c, p->neighbour[i]);
      }
    }
  }

  return u;
}

//number of useful chunks owned by a peer
static int num_chunks(struct peer *p, int t)
{
  int i, tot;

  tot = 0;
  for (i = 0; i < p->num_chunks; i++) {
    /*if (p->chunks[i] || p->recv_chunks[i]) tot++;*/
    if (!comp(i)) tot += (p->chunks[i] != NULL);
    if (!comp(i)) tot += (p->recv_chunks[i] != NULL);
  }

  return tot;
}

static void order_chunks(struct chunk_instance *c[], int *ordered, int num_chunks)
{
  int i;

  //for (i = 0; i < num_chunks; i++) chunks[i].key = (i + 1) * (chunks[i].received + 1);
  //for (i = 0; i < num_chunks; i++) chunks[i].key = i + chunks[i].received;
  for (i = 0; i < num_chunks; i++) {
    if (c[i]) {
      c[i]->key = /*i + chunks[c[i]->chunk].received */ c[i]->dl;
      //c[i]->key = i + chunks[c[i]->chunk].received;
      //c[i]->key = num_chunks - i;
    }
  }
  
  for (i = 0; i < num_chunks; i++) {
    int j, k, min;

    min = INT_MAX; k = -1;
    for (j = 0; j < num_chunks; j++) {
      if (c[j] && (c[j]->key >= 0) && (c[j]->key < min)) {
        min = c[j]->key; k = j;
      }
    }
    ordered[i] = k;
    if (k != -1) {
      c[k]->key = -1;
    }
  }
}

int is_useful(int c, struct peer *p)
{
  int i;

  for (i = 0; i < p->neigh_size; i++) {
    if (needs(c, p->neighbour[i])) {
      return 1;
    }
  }

  return 0;
}

///==========================================================
///     chunk selection algorithms
///==========================================================

SCHED int ordered_sched(struct peer *p, int t, struct peer *dummy)
{
  int i, ii;
  int ordered_chunks[p->num_chunks];

  order_chunks(p->chunks, ordered_chunks, p->num_chunks);
  for (ii = win_left(p, t); ii < win_right(p, t); ii++) {
    i = ordered_chunks[ii];
    if ((i >= 0) && p->chunks[i] && is_useful(i, p)) {
      return i;
    }
  }

  /* No useful chunk... */
  return -1;
}

//latest useful chunk
SCHED int latest_useful_chunk(struct peer *p, int t, struct peer *target)
{
  int c, j;
  for (c = win_right(p, t) - 1; c >= win_left(p, t); c--) {    //c:chunk to send
    if (p->chunks[win_idx(p, c)] && in_buffer(p, c, t)) {   
      for (j = 0; j < p->neigh_size; j++) { //j:destination node; TODO: neighbour set
        if (needs(c, p->neighbour[j])) {
          return c;
	}
      }
    }
  }
  //we don't have a useful chunk
  return -1;
}

//peer already selected, latest useful chunk
SCHED int dest_latest_useful_chunk(struct peer *p, int t, struct peer *target)
{
  int c;
  for (c = win_right(p, t) - 1; c >= win_left(p, t); c--) {    //c:chunk to send
    if (p->chunks[win_idx(p, c)] && in_buffer(p, c, t)) {   
      if (needs(c, target)) {
          return c;
      }
    }
  }
  //we don't have a useful chunk
  return -1;
}


//latest blind chunk
SCHED int latest_blind_chunk(struct peer *p, int t, struct peer *target)
{
  int c;
  for (c = win_right(p, t) - 1; c >= win_left(p, t); c--) {    //c:chunk to send
    if (p->chunks[win_idx(p, c)] && in_buffer(p, c, t)) {   
      return c;
    }
  }
  //we don't have a chunk
  return -1;
}

//earliest useful chunk
SCHED int earliest_useful_chunk(struct peer *p, int t)
{
  int c, j;
  for (c = win_left(p, t); c < win_right(p, t); c++) {    //c:chunk to send
    if (p->chunks[win_idx(p, c)] && in_buffer(p, c, t)) {   
      for (j = 0; j < p->neigh_size; j++) { //j:destination node; TODO: neighbour set
        if (needs(c, p->neighbour[j])) {
          return c;
	}
      }
    }
  }
  //we don't have a useful chunk
  return -1;
}

//peer already selected, earliest useful chunk
SCHED int dest_earliest_useful_chunk(struct peer *p, int t, struct peer *target)
{
  int c;
  for (c = win_left(p, t); c < win_right(p, t); c++) {    //c:chunk to send
    if (p->chunks[win_idx(p, c)] && in_buffer(p, c, t)) {   
      if (needs(c, target)) {
          return c;
      }
    }
  }
  //we don't have a useful chunk
  return -1;
}

//peer already selected, random useful chunk
SCHED int random_useful_chunk(struct peer *p, int t, struct peer *target)
{
  int c;
  int useful[p->num_chunks];
  int usefuls = 0;
  
  for (c = win_left(p, t); c < win_right(p, t); c++) {    //c:chunk to send
    if (p->chunks[win_idx(p, c)] && in_buffer(p, c, t)) {
      if (is_useful(c, p)) {
        useful[usefuls++] = c;
      }
    }
  }
  if (!usefuls) {
    return -1;
  } else {
    return useful[(int) (usefuls * (rand() / (RAND_MAX + 1.0)))];
  }
}

SCHED int dest_random_useful_chunk(struct peer *p, int t, struct peer *target)
{
  int c;
  int useful[p->num_chunks];
  int usefuls = 0;
  
  for (c = win_left(p, t); c < win_right(p, t); c++) {    //c:chunk to send
    if (p->chunks[win_idx(p, c)] && in_buffer(p, c, t)) {
      if (needs(c, target)) {
        useful[usefuls++] = c;
      }
    }
  }
  if (!usefuls) {
    return -1;
  } else {
    return useful[(int) (usefuls * (rand() / (RAND_MAX + 1.0)))];
  }
}

///==========================================================
///     peer selection algorithms
///==========================================================

//random peer
SCHED struct peer *random_peer(struct peer *p, int t)
{
  return p->neighbour[(int)(p->neigh_size * (rand() / (RAND_MAX + 1.0)))];
}

//most deprived peer
SCHED struct peer *most_deprived_peer(struct peer *p, int t)
{
  int n, k, i;
  n = p->num_chunks * p->neigh_size + 1; k = -1;
  
  for (i = 0; i < p->neigh_size; i++) {
    if (num_chunks(p->neighbour[i], t) < n) {
      n = num_chunks(p->neighbour[i], t);
      k = i;
    }
  }
  
  return p->neighbour[k];
}

//most deprived peer
SCHED struct peer *most_deprived_peer_random(struct peer *p, int t)
{
  int n, k, i, j, r;
  n = p->num_chunks * p->neigh_size + 1; k = -1; j = 0;
  
  for (i = 0; i < p->neigh_size; i++) {
    if (num_chunks(p->neighbour[i], t) == n) j++; //j: number of peers on being most deprived
    if (num_chunks(p->neighbour[i], t) < n) {
      n = num_chunks(p->neighbour[i], t);
      k = i;
      j = 1;
    }
  }
  
  r = (int) (j * (rand() / (RAND_MAX + 1.0)));  //r: random choice

  for (i = 0, j = 0; i < p->neigh_size; i++) {
    if (num_chunks(p->neighbour[i], t) == n && j++ == r ) {
      return p->neighbour[i];
    }
  }
  exit(-1);
}

//chunk already selected, peer that needs it
SCHED struct peer *ch_useful_peer(struct peer *p, int t, int c)
{
  int i;
  
  for (i = 0; i < p->neigh_size; i++) {
    if (needs(c, p->neighbour[i])) {
      return p->neighbour[i];
    }
  }
  //fprintf(stderr, "Warning: chunk %d was selected but no neighbours of %d need it in time %d!\n", c, p, t);
  //exit(-1);
  return NULL;
}

//chunk already selected, random choice between peers that need it
SCHED struct peer *ch_useful_peer_random(struct peer *p, int t, int c)
{
  int i, j, r;
    
  for (i = 0, j = 0; i < p->neigh_size; i++) {
    if (needs(c, p->neighbour[i])) {
      j++;
    }
  }
  
  if (!j) {
    //fprintf(stderr, "Warning: chunk %d was selected but no neighbours of %d need it in time %d!\n", c, p, t);
    //exit(-1);
    return NULL;
  }
  
  r = (int) (j * (rand() / (RAND_MAX + 1.0)));  //r: random choice

  for (i = 0, j = 0; i < p->neigh_size; i++) {
    if (needs(c, p->neighbour[i]) && j++ == r ) {
      return p->neighbour[i];
    }
  }

  return NULL;
}

//chunk already selected, most deprived peer that needs it
SCHED struct peer *ch_most_deprived_peer(struct peer *p, int t, int c)
{
  int n, k, i;
  n = p->num_chunks * p->neigh_size + 1; k = -1;
  
  for (i = 0; i < p->neigh_size; i++) {
    if (needs(c, p->neighbour[i]) && num_chunks(p->neighbour[i], t) < n) {
      n = num_chunks(p->neighbour[i], t);
      k = i;
    }
  }
  if ( k==-1 ){
    //fprintf(stderr, "Warning: chunk %d was selected but no neighbours of %d need it in time %d!\n", c, p, t);
    //exit(-1);
  }
  return p->neighbour[k];
}

//chunk already selected, random choice between most deprived peers that need it
SCHED struct peer *ch_most_deprived_peer_random(struct peer *p, int t, int c)
{
  int n, i, j, r;
  n = p->num_chunks * p->neigh_size + 1;
  
  for (i = 0, j = 0; i < p->neigh_size; i++) {
    if (needs(c, p->neighbour[i]) && num_chunks(p->neighbour[i], t) == n) j++;
    if (needs(c, p->neighbour[i]) && num_chunks(p->neighbour[i], t) < n) {
      n = num_chunks(p->neighbour[i], t);
      j = 1;
    }
  }
  
  if ( j==0 ){
    //fprintf(stderr, "Warning: chunk %d was selected but no neighbours of %d need it in time %d!\n", c, p, t);
    //exit(-1);
    return NULL;
  }

  r = (int) (j * (rand() / (RAND_MAX + 1.0)));  //r: random choice

  for (i = 0, j = 0; i < p->neigh_size; i++) {
    if (needs(c, p->neighbour[i]) && num_chunks(p->neighbour[i], t) == n && j++ == r ) {
      return p->neighbour[i];
    }
  }
  
  return NULL;
}

//node that has nothing to do in the next turn
SCHED struct peer *free_peer(struct peer *p, int t)
{
  int n, j, busy;
  
  for (n = 0; n < p->neigh_size; n++) {
    busy = 0;
    for (j = min(t, p->num_chunks - 1); j >= max(t-PO_DELAY+2,0); j--) {
      //fprintf(trace,"\ttime %d: peer %d needs chunk %d? %d\n", t, n, j, needs(n,j));
      if (!needs(j,p->neighbour[n])) {
        //fprintf(trace,"\ttime %d: peer %d busy because of chunk %d\n", t, n, j);
        busy = 1;
        break;
      }
    }
    if (!busy) {
      return p->neighbour[n];
    }
  }
  return NULL;
}

//node that has nothing to do in the next turn
SCHED struct peer *ch_almost_free_peer(struct peer *p, int t, int c)
{
  int n;

  //send to the one who has nothing to send in the next turn (t+1)
  for (n = 0; n < p->neigh_size; n++) {
    if (needs (c,p->neighbour[n]) && latest_not_needed(p->neighbour[n], t) < max(t-PO_DELAY+2,0)) {
        if (trace) fprintf(trace,"\tPeer %d: (chunk %d,%d) ---> %d, reason: node has nothing to send \n", p->id, c, c-t, n);
        return p->neighbour[n];
    }
  }

  //has nothing to send in the turn after the next one (t+2)
  for (n = 0; n < p->neigh_size; n++) {
    if (needs (c,p->neighbour[n]) && latest_not_needed(p->neighbour[n],t) < max(t-PO_DELAY+3,0)) { 
        if (trace) fprintf(trace,"\tPeer %d: (chunk %d,%d) ---> %d, reason: node has nothing to send in t+2 \n", p->id, c, c-t, n);
        return p->neighbour[n];
    }
  }

  // send to anyone who needs it
  for (n = 0; n < p->neigh_size; n++) {
    if (needs(c,p->neighbour[n])) {
      if (trace) fprintf(trace,"\tPeer %d: (chunk %d,%d) ---> %d, reason: send anyway \n", p->id, c, c-t, n);
      return p->neighbour[n];
    }
  }

  return NULL;
}

SCHED struct peer *least_utility(struct peer *p, int t, int c)
{
  int n = INT_MAX, winner = -1;
  int i;

  for (i = 0; i < p->neigh_size; i++) {
    if (needs(c, p->neighbour[i])) {
      int u = utility(p->neighbour[i], t);
      if (u < n) {
        n = u;
        winner = i;
      }
    }
  }

  if (winner >= 0) {
    return p->neighbour[winner];
  }

  return NULL;
}

///==========================================================
///     peer & chunk selection algorithms
///==========================================================

//node: destination peer & chunk selection algorithm at peer p (push)
void peer_send_c_p(struct peer *p, int t, int *chunk, struct peer **target)
{
  *chunk = chunk_sched(p, t, NULL);
  if (*chunk < 0) {
    if (trace) fprintf(trace, "\tPeer %d: Idle (no chunk to send)\n", p->id);
    *target = NULL;

    return;
  }

  *target = peer_sched(p, t, *chunk);
  if (*target == NULL) {
    if (trace) fprintf(trace, "\tPeer %d: Idle (no peer that needs chunk %d)\n", p->id, *chunk);
    return;
  }
  
  if (trace) fprintf(trace, "\tPeer %d: %d ---> %d\n", p->id, *chunk, (*target)->id);
}

//node: destination peer & chunk selection algorithm at peer p (push)
void peer_send_p_c(struct peer *p, int t, int *chunk, struct peer **target)
{
  *target = peer_sched(p, t, t);
  if (*target < 0) {
    if (trace) fprintf(trace, "\tPeer %d: Idle (no peer to send to)\n", p->id);
    return;
  }

  *chunk = chunk_sched(p, t, *target);
  if (*chunk < 0) {
    if (trace) fprintf(trace, "\tPeer %d: Idle (no chunk that %d needs)\n", p->id, (*target)->id);
    *target = NULL;
    return;
  }
  
  if (trace) fprintf(trace, "\tPeer %d: %d ---> %d\n", p->id, *chunk, (*target)->id);
}

/* Should be probably merged with peer_send_c_p */
void peer_send_utility(struct peer *p, int t, int *chunk, struct peer **target)
{
  int i, ii;
  int ordered_chunks[p->num_chunks];

  *target = NULL;
  order_chunks(p->chunks, ordered_chunks, p->num_chunks);
  for (ii = 0; ii < p->num_chunks; ii++) {
    i = ordered_chunks[ii];
    if ((i >= 0) && p->chunks[i]) {
      *target = least_utility(p, t, i);
      if (*target != NULL) {
        if (trace) fprintf(trace, "\tPeer %d: %d ---> %d\n", p->id, i, (*target)->id);
        *chunk = i;
        break;
      } else {
      }
    }
  }

  if (*target == NULL) {
    if (trace) fprintf(trace, "\tPeer %d: Idle\n", p->id);
  }
}


///==========================================================
///     source: chunk & peer selection algorithms
///==========================================================
//source: destination peer & chunk selection algorithm
//latest chunk, generic (with 3 parameter dest case "ch_")
void source_send(int t, struct peer **peer)
{
  *peer = NULL;
  if (t >= source.num_chunks) {
    return;
  }

  *peer = peer_sched(&source, t, t);        //use N as source's peer ID. hope it works

  if (trace) fprintf(trace, "\tSource: %d ---> %d\n", t, (*peer)->id);
}

void source_set(struct peer **neigh, int size, int num_chunks)
{
  source.neighbour = neigh;
  source.neigh_size = size;
  source.id = -1;
  source.num_chunks = num_chunks;
}
